# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Plugin module to find whether a large discharge is required for an ADM centrifuge """


__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


import gc
import json
import logging

from foglamp.common import logger
import filter_ingest

from foglamp.plugins.filter.ADM_LD_prediction.constants import (
    DEFAULT_TS_COL, DEFAULT_INPUTS
)
from foglamp.common.bucket_client.bucket_client import BucketClient
from foglamp.plugins.filter.ADM_LD_prediction.bucket_storage_integration import get_model_files
from foglamp.plugins.filter.ADM_LD_prediction.core import parse_config
from foglamp.plugins.filter.ADM_LD_prediction.model_inference_handler import ModelInferenceHandler
from foglamp.plugins.filter.ADM_LD_prediction.bucket_storage_integration import (
    ae_bucket_config_properties, ae_bucket_config_default,
)
_LOGGER = logger.setup(__name__, level=logging.DEBUG)


_DEFAULT_CONFIG = {
    'plugin': {   # mandatory filter
        'description': 'filter plugin to detect need for large discharge',
        'type': 'string',
        'default': 'ADM_LD_prediction',
        'readonly': 'true'
    },
    'enable': {    # recommended filter
        'description': 'Enable LD prediction plugin',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enabled',
        'order': "1"
    },
    'forwardData': {
        'description': 'Forward received vibration data down the filter chain',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Forward data',
        'order': "2"
    },
    'logLevel': {
        'description': 'log level for diagnostic or error output',
        'type': 'enumeration',
        'options': ['ERROR', 'WARN', 'INFO', 'DEBUG'],
        'default': 'INFO',
        'displayName': 'Log level',
        'order': "3"
    },
    'timeseriesAsset': {
        'description': 'Name of asset that has the vibration time series',
        'type': 'string',
        'default': "vibration",
        'displayName': 'Asset from which vibration data is read',
        'order': "4"
    },
    'centrifugeHealthAsset': {
        'description': 'Name of asset that indicates centrifuge health',
        'type': 'string',
        'default': "centrifugeHealth",
        'displayName': 'Output asset indicating centrifuge health',
        'order': "5"
    },
    'tsCol': {
        'description': 'Enter column name which contain time of the reading',
        'type': 'string',
        'default': DEFAULT_TS_COL,
        'displayName': 'Enter ts column name',
        'order': "6"
    },
    'useCols': {
        'description': 'Comma separated names of data channels without blankspace',
        'type': 'string',
        'default': ",".join(DEFAULT_INPUTS),
        'displayName': 'Comma separated Data channel(s)',
        'order': "7"
    },
    'samplingRate': {
        'description': 'Readings per sec',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sampling rate (per sec)',
        'order': "8"
    },
    'modelType': {
        'description': 'Model type to use for predicting need for large discharge',
        'type': 'enumeration',
        'default': 'Autoencoder',
        'options': ['Autoencoder'],
        'displayName': 'Model type',
        'order': '9'
    },
    'enableTpu': {
        'description': 'Connect the Coral Edge TPU and enable this',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enable Edge TPU',
        'order': '10',
        'validity': "modelType == 'Autoencoder'",
        'readonly': 'true'            # no visibility on gui
    },
    'evaluationInterval': {
        'description': 'Time interval after which to evaluate vibration data through the model '
                       '(time interval strings must be supported by pandas.to_timedelta API '
                       'e.g. 15s, 30m, 2hr, 1day 4hr 30m 10s)',
        'type': 'string',
        'default': '30min',
        'displayName': 'Vibration data evaluation interval',
        'order': "11"
    },
    'numEvaluations': {
        'description': 'Number of consecutive evaluations to consider at above specified interval',
        'type': 'integer',
        'default': '10',
        'displayName': 'Number of consecutive evaluations to consider (set)',
        'order': "12"
    },
    'numAnomalies': {
        'description': 'Number of anomalous evaluations in total number of evaluations set above',
        'type': 'integer',
        'default': '8',
        'displayName': 'Number of anomalous evaluations in set to declare anomaly',
        'order': "13"
    },
    "bucketConfig": {
        "description": "A compressed file that contains the Scaler model, ML model and Metadata.",
        "displayName": "ML Artifacts",
        "order": "14",
        "type": "bucket",
        "properties": ae_bucket_config_properties,
        "default": json.dumps(ae_bucket_config_default)
    },
}


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("plugin_info called")
    return {
        'name': 'ADM_LD_prediction',
        'version': '3.0.0-2-gabb2f04',
        'mode': 'none',
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
        ingest_ref: reference object for ingesting readings
        callback: callback to call to ingest readings to C south service
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    handle = {
        "forward_data": True,
        "enabled": False,
        "result": 0,
        "last_result": 0,
        "inference_obj": None,
        "ingestRef": ingest_ref,
        "callback": callback
    }

    try:
        module_config = parse_config(config)

        # Update logger
        _LOGGER.setLevel(module_config["log_level"])

        handle["enabled"] = module_config["enabled"]
        handle["forward_data"] = module_config["output"]["forward_data"]
        if handle["enabled"]:

            # Initialize the bucket
            try:
                if (
                    ("bucket_client" not in config) or
                    (not isinstance(config["bucket_client"], BucketClient))
                ):
                    url = config["mgmt_client_url_base"]["value"]
                    host, port = url.split(':')
                    _LOGGER.info("urlBase={}, host={}, port={}".format(url, host, port))
                    handle["mgmt_client_url_base"] = url
                    handle["bucket_client"] = BucketClient(host, port)
                else:
                    _LOGGER.info("BucketClient is already instantiated")
                    handle["mgmt_client_url_base"] = config["mgmt_client_url_base"]
                    handle["bucket_client"] = config["bucket_client"]
            except Exception as ex:
                _msg = "ADM-LD-Prediction-Filter is unable to connect to Bucket service, "
                _msg += f"restart plugin after it becomes available. Details: {str(ex)}"
                _LOGGER.error(_msg)
                raise

            # Load the Artifacts
            bucket_config = module_config["inference"]["bucket"]
            artifacts_dir = get_model_files(
                logger=_LOGGER,
                bucket_client = handle["bucket_client"],
                **bucket_config
            )

            # Initialize the plugin
            handle["inference_obj"] = ModelInferenceHandler()
            _ = handle["inference_obj"].initialize(
                module_config, artifacts_dir, _LOGGER
            )

            _LOGGER.info("ADM LD Prediction Filter has been initialized properly.")

    except Exception as ex:
        _LOGGER.exception('Error in initializing ADM_LD_prediction plugin {}'.format(str(ex)))
        raise

    return handle


def plugin_ingest(handle, data):
    """ Returns assets to monitor for centrifuge vibration data
    Args:
        handle: handle returned by the plugin initialization call
        data: list of
        dicts of
          'readings': {'Channel1': v1, 'Channel2': v2, 'Channel3': v3, 'Channel4': v4, 'user_ts': timestamp},
          'asset': asset-name,
          'id': id,
          'ts': current-timestamp,
          'user_ts': timestamp-of-data
    Action:
        passes data to analysis algorithm, which returns:
        0: LD not required
        1: LD required
        calls the ingest callback to pass along input data
        calls the ingest callback with an asset that holds the current LD required state
    Returns:
    """
    if isinstance(data, dict):
        data = [data]

    # Check if execution allowed
    if handle["inference_obj"] is None or len(data) == 0:
        return

    # if filter is disabled and data forwarding is allowed then pass on the data without modification
    if not handle["enabled"] and handle["forward_data"]:
        _LOGGER.debug("Forwarding original {} readings".format(len(data)))
        filter_ingest.filter_ingest_callback(handle["callback"], handle["ingestRef"], data)
        del data
        return

    # perform data ingestion & anomaly detection
    result, data_to_fwd = handle["inference_obj"].run(data, handle["last_result"])
    if result >= 0:  # only valid result; `-1` can represent performing inference was not allowed
        handle["last_result"] = result

    filter_ingest.filter_ingest_callback(handle["callback"], handle["ingestRef"], data_to_fwd)

    del data, data_to_fwd, result
    gc.collect()


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialization call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    if (
        ("mgmt_client_url_base" in handle) and ("bucket_client" in handle)
    ):
        new_config["mgmt_client_url_base"] = handle["mgmt_client_url_base"]
        new_config["bucket_client"] = handle["bucket_client"]

    new_handle = plugin_init(new_config, handle["ingestRef"], handle["callback"])
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
    Args:
        handle: handle returned by the plugin initialization call
    Returns:
        plugin shutdown
    """
    _LOGGER.info('ADM_LD_prediction plugin shut down.')

    if handle["enabled"]:
        handle["inference_obj"].shutdown()

    handle["inference_obj"] = None
    _LOGGER.info("ADM_LD_prediction::plugin_shutdown() completed")
