"""A file containing the config and code for interacting with Bucket Service."""

import os
import glob
import asyncio
import datetime
import tarfile
import joblib
import pickle
import numpy as np

from foglamp.common.bucket_client.model_query_builder import ModelQueryBuilder

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


ae_bucket_config_properties = {
    "constant": {
        "type": "model"
    },
    "key": {
        "name": {
            "description": "Archive containing models that is to be picked from bucket service.",
            "type": "string",
            "default": "Model archive",
            "order": "1",
            "displayName": "Model archive name"
        },
    },
    "properties": {
        "version": {
            "description": "Archive version as stored in bucket",
            "type": "string",
            "default": "1.0",
            "order": "2",
            "displayName": "Model version"
        },
        "scaler": {
            "description": "The file name of scaler that will be loaded from the archive.",
            "type": "string",
            "default": "ae_min_max_scaler.pkl",
            "order": "3",
            "displayName": "Scaler",
            "readonly": "true"
        },
        "ml_model": {
            "description": "The file name of ML model that will be loaded from the archive.",
            "type": "string",
            "default": "adm_ld_ae_float32.tflite",
            "order": "4",
            "displayName": "ML model",
            "readonly": "true"
        },
        "model_metadata": {
            "description": "The file name of model metadata that will be loaded from the archive.",
            "type": "string",
            "default": "ae_model_data.npy",
            "order": "5",
            "displayName": "Model metadata",
            "readonly": "true"
        }
    }
}

ae_bucket_config_default = {
    "type": ae_bucket_config_properties["constant"]["type"],
    "name": ae_bucket_config_properties["key"]["name"]["default"],
    "version": ae_bucket_config_properties["properties"]["version"]["default"],
    "scaler": ae_bucket_config_properties["properties"]["scaler"]["default"],
    "ml_model": ae_bucket_config_properties["properties"]["ml_model"]["default"],
    "model_metadata": ae_bucket_config_properties["properties"]["model_metadata"]["default"],
}


def get_model_files(
    logger, bucket_client, bucket_type, model_name, model_version
):
    """Get model archive file from the bucket service and extract it at a path.

    FUTURE: how about loading the model directly without writing to disk
    Setting up paths
        - model_dir directory: where the artifacts will be extracted
        - archive saving path: where archive coming from bucket service will be
                               saved

    Args:
        bucket_client: Bucket client handle to use bucket services
        bucket_type: type defined in the attributes in bucket.
            it should match with
            ae_bucket_config_properties.constant.type --> "adm_ld_filter_models"
        model_name: Model name e.g. "TFLite AE Model Archive"
        model_version: Model's version e.g. "1.0"
    Returns:
        Model directory: where the models were extracted.
    Raises:
    """
    # Paths setup
    model_dir = os.path.join("/tmp", "foglamp_adm_ld_filter")
    archive_saving_path = os.path.join(model_dir, "adm_ld_filter_archive.tar.gz")

    # Prepare the directory
    os.system(f"rm -rf {model_dir}")
    os.makedirs(model_dir)

    # creating query for searching the bucket using the attributes
    model_attrs = {
        "type": bucket_type,
        "name": model_name,
        "version": model_version,
    }
    mqb = ModelQueryBuilder()
    for param, user_defined_value in model_attrs.items():
        mqb = mqb.MATCH([param, "=", user_defined_value])
    query = mqb.payload()
    logger.info(f"bucket_client.query_models: query payload={query}")

    # Create a temporary event loop for async calls in this function since
    # sometimes this function gets called before event loop creation
    # (plugin_init) and sometimes after (plugin_reconfigure)
    loop = asyncio.new_event_loop()

    # Locating the matching bucket
    try:
        # getting the matching bucket options based on the query created above
        task = asyncio.ensure_future(
            bucket_client.query_models(query), loop=loop
        )
        loop.run_until_complete(task)
        result = task.result()
        logger.info(f"bucket_client.query_models returned {result}")

        # checking the matches
        if len(result["matches"]) < 1:
            logger.error("Didn\'t find any matching data within bucket service.")
            return None
        if len(result["matches"]) == 1:
            logger.info("Found the exact match within bucket service. Loading it.")
        if len(result["matches"]) > 1:
            logger.warn("Found more than one match within bucket service. Picking first one.")

        # loading the match
        result = result["matches"][0]
        logger.info(f"Chosen model : {result}")

    except Exception as ex:
        _msg = "ADM LD filter is unable to get a match within bucket service."
        _msg += f"\nDetails: {str(ex)}"
        logger.error(_msg)
        return None

    # Get the data from the bucket & save it to the mentioned path
    task = asyncio.ensure_future(
        bucket_client.get_bucket_content(result["bucket"], archive_saving_path),
        loop=loop
    )
    loop.run_until_complete(task)
    result = task.result()
    loop.stop()
    loop.close()
    logger.info(f"Saved the archive from the bucket to {archive_saving_path}")

    # Extract the archive to the model directory
    logger.info(f"Un-compressing archive {archive_saving_path} in {model_dir}")
    try:
        tar = tarfile.open(archive_saving_path, "r|*")
        tar.extractall(model_dir)
        tar.close()
    except Exception as ex:
        logger.error(
            "Unable to extract model files from (assumed) compressed archive:" +
            f"{result['file']}, exception={str(ex)}"
        )
        return None

    return model_dir


def load_module_artifacts(logger, model_dir, model_type, use_tpu=False):
    """Load the module artifacts according to extension and provided config."""
    if model_dir is None:
        raise Exception("Unable to get models from the bucket")

    # set up the various model files used to do ML/NN/AE analysis
    scaler_paths = sorted(glob.glob(os.path.join(model_dir, "*.pkl")))
    model_paths = sorted(glob.glob(os.path.join(model_dir, "*.tflite")))
    metadata_paths = sorted(glob.glob(os.path.join(model_dir, "*.npy")))

    scaler_path = scaler_paths[0] if len(scaler_paths) > 0 else ""
    metadata_path = metadata_paths[0] if len(metadata_paths) > 0 else ""
    model_path = model_paths[0] if len(model_paths) > 0 else ""

    _msg = "\nModel will be loaded from\n\t1.Scaler: {}\n\t2.Model: {}\n\t3.Metadata: {}\n".format(
        scaler_path, model_path, metadata_path
    )
    logger.info(_msg)

    for path in [scaler_path, model_path, metadata_path]:
        if not os.path.isfile(path):
            raise Exception("Unable to find file: {}\n Please restart the corresponding service.".format(path))
        else:
            logger.debug("Found file: {}".format(path))
            pass

    # Load artifacts
    min_max_scaler = pickle.load(open(scaler_path, "rb"))
    model_metadata = np.load(metadata_path, allow_pickle=True).item()

    if model_type == "Xgboost":
        model = joblib.load(model_path)

    elif model_type in ["Neural network", "Autoencoder"]:

        # create the tensorflow runtime,
        try:
            from tflite_runtime.interpreter import Interpreter

            try:
                from tflite_runtime.interpreter import load_delegate
            except ImportError:
                logger.error("Edge TPU Support is not found in Tensorflow installation ")
            except:
                logger.error("TFlite runtime installation not found")
                raise

            t1 = datetime.datetime.now()

            if use_tpu:
                # load the Edge TPU Runtime
                model = Interpreter(
                    model_path=model_path,
                    experimental_delegates=[load_delegate("libedgetpu.so.1.0")]
                )
            else:
                model = Interpreter(model_path=model_path)

            model.allocate_tensors()
            model = model

            t2 = datetime.datetime.now()
            delta = t2 - t1
            logger.info("Neural network model loaded in {:.2f} msec".format(delta.total_seconds() * 1000))
        except Exception as e:
            model = None
            logger.error(
                'Failed to create TFlite interpreter instance: {}{}'
                .format(
                    e, "   Check whether edge TPU is attached" if use_tpu else ""
                )
            )
            raise

    return min_max_scaler, model, model_metadata
