# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Server functions for plugin module to detect need for large discharge (LD) """

__author__ = "Amandeep Singh Arora, Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


import datetime
import numpy as np

from foglamp.plugins.filter.ADM_LD_prediction import wpt_features as wpt


def adjust_CPT5_columns(X):
    """ adjust_CPT5_columns - adjust CPT5 column values in lowest frequency node of WPD
        Arguments:
            X - input feature data
        Returns:
            processed feature data
    """
    X[:,320*0+16] = np.log(X[:,320*0+16])
    X[:,320*1+16] = np.log(X[:,320*1+16])
    X[:,320*2+16] = np.log(X[:,320*2+16])
    X[:,320*3+16] = np.log(X[:,320*3+16])

    return X


def check_anomaly(self):
    """ check_anomaly - run model inference to look for anomaly in time series data
        Arguments:

        Returns:
            y_pred - array of predictive values; all zeros indicate a perfect match for expected data
                     all ones indicate a perfect match for anomalous data
            RE_list - for each entry above, the reconstruction error (RE) to RE threshold ratio values
            smoothed_RE_list - for each y_pred entry above, the reconstruction error (RE) to RE threshold ratio values smoothed using a savgol filter
    """

    t1 = datetime.datetime.now()

    if (self.model_type in ['Neural network', 'Autoencoder']) and self.model is None:
        # sanity check
        self.logger.error("Neural net interpreter not present")
        raise ValueError

    try:
        interval_rows = int(self.frame_size*1)

        # extract features from time series data using wavelet packet decomposition
        features = wpt.parse_vibration_data(
            self.frame, self.sampling_rate, interval_rows, 4
        )
        features2 = np.array(features)
        n_nan_features = np.count_nonzero(np.isnan(features2), axis=0).sum()
        if n_nan_features > 0:
            self.logger.warn(
                f"The input value of features are observed to contain NaNs with"
                " `{n_nan_features}` features out of `{features2.shape[1]}` "
                "features."
            )

        n_inf_features = np.count_nonzero(np.isinf(features2), axis=0).sum()
        if n_inf_features > 0:
            self.logger.warn(
                f"The input value of features are observed to contain Infinity "
                "with `{n_inf_features}` features out of `{features2.shape[1]}`"
                " features."
            )

        # Skip inference if presence of NaN or Inf is observed in input
        if ((n_nan_features > 0) or (n_inf_features > 0)):
            self.logger.error("Skipping inference cycle as NaN or Inf is observed in input.")
            return -1, [], []

        features3 = self.min_max_scaler.transform(features2)

        # -- multiple ml choices --
        # either:
        # . use xgboost to create a decision tree of features that match or don't match "healthy" centrifuge state
        # . use a DNN to classify the input as normal / not normal based on wpd training
        # . use an autoencoder DNN to decide if the output is too far from the input to be a normal mapping

        RE_list = []

        if self.model_type == 'Xgboost':
            y_pred = self.model.predict(features3)
            self.y_pred = np.append(self.y_pred, y_pred)
        elif self.model_type == 'Autoencoder':
            threshold = self.model_metadata[
                "RE_threshold_uint8_quant" if self.tpu else "RE_threshold"
            ]
            y_pred,RE_list,smoothed_RE_list = self.ae_prediction(features3, self.tpu, self.model, threshold)
            self.y_pred = np.append(self.y_pred, y_pred)
        else:
            self.logger.debug("calling nn")

            y_pred = self.nn_prediction(features3, self.tpu, self.model)
            # threshold = 0.8
            y_pred = y_pred.round()  # binarize probability to one of the 2 classes

            self.y_pred = np.append(self.y_pred, y_pred)

        if self.y_pred.shape[0] > self.max_prediction:
            to_del = self.y_pred.shape[0]-self.max_prediction
            self.y_pred = np.delete(self.y_pred, slice(0,to_del), 0)

        self.logger.debug("y_pred={} -> {} ({})".format(y_pred, self.y_pred, np.count_nonzero(self.y_pred)))
        self.logger.debug("------ self.y_pred: sum={}, length={}, ratio={}, num_frames_processed={}, num_input_rows={}, so rv={}".
                         format(self.y_pred.sum(), len(self.y_pred),
                                (self.y_pred.sum() / len(self.y_pred)).round(2),
                                self.num_rows_processed // self.frame_size,
                                self.num_input_rows,
                                np.count_nonzero(self.y_pred) >= self.min_nonzero_anomaly))

        y_pred = 1 if (np.count_nonzero(self.y_pred) >= self.min_nonzero_anomaly) else 0

    except Exception as ex:
        self.logger.error('Error in evaluating ADM LD plugin {}'.format(str(ex)))
        raise

    t2 = datetime.datetime.now()
    delta = t2 - t1
    self.logger.debug("check_anomaly took {} sec, returning y_pred={}".format(int(delta.total_seconds() * 1000 * 1000), y_pred))

    return y_pred, RE_list, smoothed_RE_list   # y_pred: 0 means LD not required, 1 means LD required


def ae_prediction(self, features4, tpu, interpreter, threshold):
    """ get prediction on features list using auto encoder
        Args:
            features4: processed WPD statistical features
            interpreter: tensorflow interpreter
            tpu: bool, use tpu or not
            threshold: Maximum reconstruction error for normal centrifuge operation
        Returns:
            Returns prediction for each feature row in the input
    """

    features4 = features4.astype('float32')

    # output = interpreter.tensor(interpreter.get_output_details()[0]["index"])
    # self.logger.info("interpreter.get_input_details()={}".format(interpreter.get_input_details()))
    # self.logger.info("interpreter.get_output_details()={}".format(interpreter.get_output_details()))

    if tpu:
        # for TPU normalize to a zero point (then reverse after creating prediction)
        v = interpreter.get_input_details()[0]['quantization']
        scale = v[0]
        zero_point = v[1]
        features4 = np.round(features4 / scale + zero_point).astype('uint8')

    t1 = datetime.datetime.now()

    y_pred = []
    mse_list = []
    smoothed_mse_list = []
    for r in range(features4.shape[0]):
        feat = np.expand_dims(features4[r], axis=0)
        # for each of the input feature row, create an output prediction
        interpreter.set_tensor(interpreter.get_input_details()[0]['index'], feat)
        interpreter.invoke()
        prediction = interpreter.get_tensor(interpreter.get_output_details()[0]['index'])
        # self.logger.info("interpreter.get_output_details()={}".format(interpreter.get_output_details()))

        if tpu:
            # reverse the normalization
            v = interpreter.get_output_details()[0]['quantization']
            scale = v[0]
            zero_point = v[1]
            prediction = (prediction.astype('float32') - zero_point) * scale

        prediction = np.array(prediction)
        mse = np.mean(np.power(feat - prediction, 2), axis=1)

        # Make sure we don't have NaN's in prediction
        mse[~np.isfinite(mse)] = 0.5*threshold # replace NaN with 0.5
        mse = np.clip(mse, 0, 5.0*threshold)

        smoothed_mse = self.smoothing_filter.update(mse[0])

        self.logger.debug("r={}, mse={}, smoothed_mse={}", r, mse, smoothed_mse)
        mse_list.append(mse[0])
        smoothed_mse_list.append(smoothed_mse)
        y_pred.append(1 if mse >= threshold else 0)

    # return the prediction -- a majority of outputs need to signal in order to be classified as anomalous
    y_pred = np.array(y_pred)
    self.logger.debug("threshold={}, mse_list={}, smoothed_mse_list={}".format(threshold, 
                                                                               [ "{:.8f}".format(mse) for mse in mse_list],
                                                                               [ "{:.8f}".format(smse) for smse in smoothed_mse_list] ))
    t2 = datetime.datetime.now()
    delta = t2 - t1
    self.logger.debug("Inference done in {} usec".format(int(delta.total_seconds() * 1000 * 1000)))

    RE_list = ["{:.4f}".format(k/threshold) for k in mse_list]
    smoothed_RE_list = ["{:.4f}".format(k/threshold) for k in smoothed_mse_list]
    return y_pred, RE_list, smoothed_RE_list

export_service_functions = [check_anomaly, ae_prediction]
