# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Server functions for plugin module to detect need for large discharge (LD) """


__author__ = "Amandeep Singh Arora, Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"

import os
import gc
import datetime
import logging

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter

from foglamp.common import logger, utils
from foglamp.plugins.filter.ADM_LD_prediction.constants import (
    LOGLEVEL, DEFAULT_TS_COL, DEFAULT_INPUTS
)
from foglamp.plugins.filter.ADM_LD_prediction import ml_eval
from foglamp.plugins.filter.ADM_LD_prediction.bucket_storage_integration import load_module_artifacts


# very special: if we are passed in a default logger destination use that
if hasattr(logger, 'FOGLAMP_DESTINATION_ENV') and (logger.FOGLAMP_DESTINATION_ENV in os.environ):
    logger.set_default_destination(int(os.environ[logger.FOGLAMP_DESTINATION_ENV]))
else:
    utils.eprint("NO DEST ENV")


class StreamingSavgolFilter:
    def __init__(self, window_size, poly_order, logger):
        self.window_size = window_size
        self.poly_order = poly_order
        self.data_window = []
        self.logger = logger

    def update(self, new_value):
        # Add the new value to the data window
        self.data_window.append(new_value)
        self.logger.debug("len(self.data_window)={}, self.window_size={}".format(len(self.data_window), self.window_size))

        # If the window is full, apply the Savitzky-Golay filter
        if len(self.data_window) >= self.window_size or len(self.data_window) > self.poly_order+1:
            smoothed_value = self.apply_savgol_filter()
            self.logger.debug(f"New Value: {new_value:.2f}, Smoothed Value: {smoothed_value:.2f}")
            return smoothed_value
        else:
            return new_value

    def apply_savgol_filter(self):
        # Apply the Savitzky-Golay filter to the data window; min valid window size is poly_order+1 and it has to be odd-sized
        window_sz = self.window_size if (len(self.data_window) > self.window_size) else ((len(self.data_window)-1)//2*2+1)
        self.logger.debug("self.window_size={}, len(self.data_window)={}, hence using window_sz={}".format(self.window_size, len(self.data_window), window_sz))
        smoothed_value = savgol_filter(self.data_window, window_sz, self.poly_order)
        return smoothed_value[-1]  # Return the last smoothed value


@utils.add_functions_as_methods(ml_eval.export_service_functions)
class ModelInferenceHandler:

    def __init__(self):
        self._default_init()

    def _default_init(self):
        self.logger = logger.setup(__name__, level=logging.DEBUG)

        # data are input into an in-memory frame
        self.df = pd.DataFrame()
        self.num_rows_processed = 0
        self.num_input_rows = 0
        self.num_frames = 0
        self.frame = None
        self.frames = []

        self.min_max_scaler = None
        self.model = None
        self.model_data = None

        self.tpu = False
        self.allow_execution = False

        self.do_check_anomaly = True
        self.sampling_rate = 8000
        self.frame_size = self.sampling_rate * 5
        self.inference_gap = None
        self.all_cols = [DEFAULT_TS_COL] + DEFAULT_INPUTS

        self.lastProcessedDataTS = pd.Timestamp('2000-01-01 00:00:00.000000+00:00')


    def initialize(self, module_config, artifacts_dir, logger=None):
        """Initialize the inference handler class.

        Override the default config if needed.
        """
        if logger is not None:
            self.logger = logger
        self.logger.debug('Inside ModelInferenceHandler.initialize()')

        # Inputs
        self.input_asset = module_config["inputs"]["asset"]
        self.sampling_rate = module_config["inputs"]["sampling_rate"]

        # Inference
        self.tpu = module_config["inference"]["use_tpu"]
        self.model_type = module_config["inference"]["model_type"]
        self.ts_col = module_config["inference"]["ts_col"]
        self.sensor_cols = module_config["inference"]["sensor_cols"]
        self.inference_gap = module_config["inference"]["trigger_interval"]

        # Post Process
        self.max_prediction = int(module_config["post_process"]["num_evaluation"])
        self.min_nonzero_anomaly = int(module_config["post_process"]["num_anomaly"])

        # Output
        self.output_asset = module_config["output"]["asset"]
        self.forward_data = module_config["output"]["forward_data"]

        # Calculated config
        self.frame_size = self.sampling_rate * 5
        self.all_cols = [self.ts_col] + self.sensor_cols

        # accumulate 0/1 from the past n frames in an array; anomaly is when m/n are 1
        self.y_pred = np.array(0)

        # window_size of 31 is reasonable balance between smoothing requirement while
        # still retaining sudden peaks/valleys without too much phase delay
        self.smoothing_filter = StreamingSavgolFilter(31, 3, self.logger)

        (
            self.min_max_scaler, self.model, self.model_metadata
        ) = load_module_artifacts(self.logger, artifacts_dir, self.model_type, self.tpu)

        self.allow_execution = True

    def shutdown(self):
        self.logger.warning("ModelServer: SHUTTING DOWN")
        self.allow_execution = False

    def evaluate(self, data):
        """ Evaluates whether a large discharge is required by running vibration data thru' ML model
        Args:
            data: 2 array of readings:
            {'channel1': [amp1, amp2, ...]
             'ts': [ts1, ts2, ...]}
        Returns:
            Returns a bool indicating whether LD is required
                - -1 means insufficient readings or presence of issue
                - 0 means LD not required
                - 1 means LD required
        """
        self.logger.debug("ModelServer: plugin_ingest")

        if not self.allow_execution:
            self.logger.error("Cannot process data since NN interpreter is not available")
            return -1, [], []

        try:
            if not self.accumulate_data(data):
                return -1, [], []

            anomalous, RE_list, smoothed_RE_list = self.check_anomaly()
            return anomalous, RE_list, smoothed_RE_list

        except Exception as ex:
            self.logger.error('Error in ModelServer plugin_ingest {}'.format(str(ex)))
            self.allow_execution = False
            raise

    def accumulate_data(self, data):
        """ accumulate_data - add data to the dataframe we assemble for analysis """

        # clean data: convert to canonical names
        if len(data) == 0:
            return []

        t1 = datetime.datetime.now()

        _input = pd.DataFrame(data[0])
        currTS = pd.to_datetime(_input['user_ts'][0])  # TS in first reading
        if currTS < self.lastProcessedDataTS + self.inference_gap:
            self.logger.debug("lastProcessedDataTS={}, inference_gap={}, currTS={}, so NOT processing this data frame"
                              .format(self.lastProcessedDataTS, self.inference_gap, currTS))

            del data, _input
            return False

        t1 = datetime.datetime.now()
        # The following code assumes that the data coming to the filter
        # is a list of dicts, where each dict is not a nested dict.
        # If nested dicts are there then use the following call.
        #  _input = _input.join(pd.json_normalize(_input['readings']))
        required_dict_list = [s['readings'] for s in data]
        required_ts_list = [s['user_ts'] for s in data]
        _input_d = pd.DataFrame(required_ts_list, columns=['user_ts_orig'])
        _input_r = pd.DataFrame.from_records(required_dict_list)
        _input = pd.concat([_input_r, _input_d], axis=1)
        extra_cols = list(_input.columns)
        extra_cols.remove('user_ts_orig')

        extra_cols = set(extra_cols) - set(self.all_cols)
        self.logger.debug(
            "self.all_cols={}, extra_cols = {}".format(self.all_cols, extra_cols)
        )
        if len(extra_cols):
            _input.drop(labels=list(extra_cols), axis=1, inplace=True)

        _input.index = pd.to_datetime(_input['user_ts_orig'])
        labels_to_remove = ['user_ts_orig']
        if 'user_ts' in _input.columns:
            labels_to_remove.append('user_ts')
        _input.drop(labels=labels_to_remove, axis=1, inplace=True)

        t2 = datetime.datetime.now()
        delta = t2 - t1
        delta = int(delta.total_seconds() * 1000 * 1000)  # in usec
        self.logger.debug(
            "df processing took {:.0f} msec for {} readings, time/reading={:.0f} us"
                .format(delta/1000, _input.shape[0], delta/_input.shape[0]))

        self.logger.debug("lastProcessedDataTS={}, inference_gap={}, currTS={}, so processing this data frame"
                          .format(self.lastProcessedDataTS, self.inference_gap, currTS))

        _len = _input.shape[0]
        self.df = pd.concat([self.df, pd.DataFrame(_input)], ignore_index=True)
        self.num_input_rows += _len

        if self.df.shape[0] < self.frame_size:
            self.logger.debug("###### Got {} rows to process, not processing yet".format(self.df.shape[0]))
            return False   # don't run model inference if data length is not sufficient

        self.lastProcessedDataTS = currTS

        self.num_rows_processed += _len
        self.num_frames += 1

        self.frame = self.df[-self.frame_size:]

        self.df = pd.DataFrame()

        return True

    def run(self, data, last_result):
        """Execute the anomaly detection"""
        # if incoming reading list has a mix of assets, process only the matching readings
        non_matching_asset_readings = [d for d in data if d["asset"] != self.input_asset]

        matching_asset_readings = [d for d in data if d["asset"] == self.input_asset]
        self.logger.debug("Processing {} readings with matching asset".format(len(matching_asset_readings)))

        if self.forward_data:
            data_to_fwd = data
        else:
            data_to_fwd = non_matching_asset_readings

        # run model and check need for large discharge
        result, RE_list, smoothed_RE_list = self.evaluate(matching_asset_readings)
        self.logger.debug(
            "plugin_ingest (RPC call) returned: _result={}, type {};  RE_list={}, type={} "
            .format(result, type(result), RE_list, type(RE_list))
        )
        if len(RE_list) > 1:
            self.logger.warn("RE_list has more than 1 element, probably unhandled case")

        if result >= 0:  # valid

            if result != last_result:
                self.logger.info("ingest: state change {} -> {} ".format(last_result, result))

            # must have 1-1 assets we add vs assets we receive
            if "timestamp" in matching_asset_readings[-1]:
                timestamp = matching_asset_readings[-1]["timestamp"]
            else:
                timestamp = matching_asset_readings[-1]["ts"]

            output_dict = {
                "asset": self.output_asset,
                "readings": {
                    "LD_required": result,
                    "RE_threshold_multiple": float(smoothed_RE_list[0]),
                    "RE_threshold_multiple_raw": float(RE_list[0])
                },
                "timestamp": timestamp
            }

            data_to_fwd.append(output_dict)

            # pass new asset status data up to upper levels
            self.logger.info("send health asset: {}".format([output_dict]))

        del data, matching_asset_readings, non_matching_asset_readings
        gc.collect()
        return result, data_to_fwd
